﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Encodings;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;

namespace HeuristicLab.Problems.VehicleRouting {
  /// <summary>
  /// An operator that relinks paths between vehicle routing solutions.
  /// </summary>
  [Item("VRPPathRelinker", "An operator that relinks paths between vehicle routing solutions.")]
  [StorableClass]
  public sealed class VRPPathRelinker : PathRelinker {
    [StorableConstructor]
    private VRPPathRelinker(bool deserializing) : base(deserializing) { }
    private VRPPathRelinker(VRPPathRelinker original, Cloner cloner) : base(original, cloner) { }
    public VRPPathRelinker() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new VRPPathRelinker(this, cloner);
    }

    public static ItemArray<IItem> Apply(IItem initiator, IItem guide, PercentValue n) {
      PotvinEncoding sol1 = initiator.Clone() as PotvinEncoding;
      PotvinEncoding sol2 = guide.Clone() as PotvinEncoding;
      PotvinEncoding sol3 = new PotvinEncoding();

      // match tours
      foreach (Tour tour1 in sol1.Tours) {
        int highestMatch = -1;
        Tour bestMatchingTour = null;
        foreach (Tour tour2 in sol2.Tours) {
          int matchingCities = MatchingCities(tour1, tour2);
          if (matchingCities > highestMatch) {
            bestMatchingTour = tour2;
            highestMatch = matchingCities;
          }
        }
        sol2.Tours.Remove(bestMatchingTour);
        sol3.Tours.Add(bestMatchingTour);
      }
      // add all remaining tours
      sol3.Tours.AddRange(sol2.Tours);

      IList<PotvinEncoding> solutions = new List<PotvinEncoding>();
      // TODO: add local search

      IList<IItem> selection = new List<IItem>();
      if (solutions.Count > 0) {
        int noSol = (int)(solutions.Count * n.Value);
        if (noSol <= 0) noSol++;
        double stepSize = (double)solutions.Count / (double)noSol;
        for (int i = 0; i < noSol; i++)
          selection.Add(solutions.ElementAt((int)((i + 1) * stepSize - stepSize * 0.5)));
      }

      return new ItemArray<IItem>(selection);
    }

    private static int MatchingCities(Tour tour1, Tour tour2) {
      return tour1.Cities.Intersect(tour2.Cities).Count();
    }

    protected override ItemArray<IItem> Relink(ItemArray<IItem> parents, PercentValue n) {
      if (parents.Length != 2)
        throw new ArgumentException("The number of parents is not equal to 2.");
      return Apply(parents[0], parents[1], n);
    }
  }
}
